
/*
 * Copyright 2008, Russ White
 * 
 * Twisted Pear Audio Buffalo DAC firmware with volume control for PIC16F690 or PIC16F685 (I think others would work too)
 * 
 * The Firmware will work on the JT controller PCB
 * 
 * I only used this PIC because I had some on hand. :)
 * 
 * The software is provided as is and you shouldnot assume it works!
 * 
 * Use this software at your own risk... Its not my fault if it blows up anything you own. ;)
 * 
 */

#include <htc.h>
#include "delay.h"
#include "init.h"

#define POT RA0 								// ADC input
#define AUTOMUTE_TRIGGER_LOWEST 	0b1111111 	// lowest automute trigger point
#define AUTOMUTE_OFF				0b00000000	// Effectively turns off automute
#define SET_LENGTH					4			// The size of the set for averaging
#define LOW_VOLUME 					127			// Lower limit on volume, to conserve pot travel....

// external I2C functions
char I2C_Send(char Address,char *Data,char Num);
char I2C_Read(char Address,char *Data,char Num);
char i2csendbyte(char byte);
char i2cgetbyte(void);
char i2cgetack(void);
void i2csendack(void);
void i2cdelay(char delay);
void i2cstart(void);
void i2cstop(void);
void i2cclock(void);
char i2creadbit(void);

// delay
void DelayMs(unsigned char);

// local functions
unsigned char getState (void);
unsigned char getAVState (void);
void change(unsigned char);
void setRegister (unsigned char, unsigned char);
void setRegBits(int reg, int high, int low, int value);
void setup(void);


// Globals

unsigned char getState(void);

unsigned char last, value;

unsigned char address = 0x90;
// SPDIF/PCM enable
enum {DISABLE,ENABLE} SPDIF_ENABLE = ENABLE;
// PCM data length
enum {_24BIT,_20BIT,_16BIT} PCM_DATA_LENGTH = _24BIT;
// PCM data format
enum {I2S,LJ,RJ} PCM_DATA_FORMAT = I2S;
// NON-OVERLAP control
enum {OVERLAP,STANDARD} NON_OVERLAP_CONTROL = STANDARD;
// Jitter reduction enable
enum Enable {DISABLED, ENABLED} JITTER_REDUCTION = ENABLED;
// Deemphasis flag
enum Bypass {USE,BYPASS} BYPASS_DEEMPHASIS = BYPASS;
// Mute all DACs
enum {UNMUTED, MUTED} MUTE = UNMUTED;
// NS Modulator GAIN exponential part
enum {EXP_3,EXP_4} NS_MOD_GAIN = EXP_4;
// DPLL bandwidth
enum {NONE, LOWEST, MED_LOW, MEDIUM, MED_HIGH, HIGHEST} DPLL_BANDWIDTH = LOWEST;
// PCM deemphasis frequency.
enum {_32,_41_1,_48,RESERVED} PCM_DEEMPH_FREQ = _41_1;
// Dither Control
enum {NOISE_SHAPED,FIXED} DITHER_CONTROL = NOISE_SHAPED;
// Rotator input
enum {NS_MOD, External} ROTATOR_INPUT = NS_MOD;
// remapping for maximum phase seperation
enum {NO_REMAPPING, REMAP} REMMAPPING = REMAP;
// notch delay
enum {NO_NOTCH = 0b00000,DIV4 = 0b00001,DIV8 = 0b00011,DIV16 = 0b00111,DIV32 = 0b01111,DIV64 = 0b11111} NOTCH_DELAY = NO_NOTCH;
// DAC polarity
enum Polarity {IN_PHASE, ANTI_PHASE};
enum Polarity DAC_POLARITY = IN_PHASE;
// DAC source
enum DACSource {NORMAL, ALTERNATE};
enum DACSource DAC_8_7_4_3_SOURCE = ALTERNATE;
// differential mode
enum {TRUE_DIFFERENTIAL, PSEUDO_DIFFERENTIAL} DIFFERENTIAL_MODE = TRUE_DIFFERENTIAL;
// IIR Bandwidth
enum {NORM,_50K,_60K,_70K} IIR_BANDWIDTH = _50K;
// FIR roll off
enum {SLOW,FAST} ROLLOFF = SLOW;
// Quantizer size
enum {Q6,Q7,Q8,Q9} QUANTIZER_SIZE = Q6;

unsigned char registers[16];

main(void) {
	int i;

	init();
	// setup the ADC on AN0 left justified
	ANSEL = 0b001110001;
	// turn ADC on
	ADCON0 = 0b000000001;
	last = LOW_VOLUME;
	setup();
	DelayMs(50); // allow ADC and Buffalo to settle in
	for (i = 0;i < 16; i++) {
		setRegister(i,registers[i]);
	}
	// allow ADC and Buffalo to settle in
	DelayMs(250);
	DelayMs(250);
	DelayMs(250);
	DelayMs(250);
	value = getAVState();
	// do a slow ramp up.
	do {
		change(last);
		value = getAVState();
		if (last - value > 12) last -= 3;
		else last--;
	} while (last > value);
	// loop forever polling ADC.
	while (1){
		value = getAVState();
		if (value != last) {
			last = value;
			change(last);
		}
	}
}

void setRegister (unsigned char reg, unsigned char value) {
	unsigned char message[2];
	message[0] = reg;
	message[1] = value;
	I2C_Send(address,message,2);
}


unsigned char getState () {
	GODONE = 1;
	while (GODONE){
		NOP();
	}
	DelayUs(10);
	return LOW_VOLUME - (ADRESH >> 1);
}

unsigned char getAVState () {
	int sum = 0;
	int i;
	for (i = 0; i < SET_LENGTH; i++) {
		//Let the ADC recover...
		if (i < (SET_LENGTH -1)) DelayMs(8);
		sum += getState();
	}
	return (unsigned char) (sum / SET_LENGTH);
}

void change(unsigned char val) {
	int i;
	for (i = 0; i < 8; i++)	setRegister(i,val);
}

void setup() {
	registers[0] = LOW_VOLUME;
	registers[1] = LOW_VOLUME;
	registers[2] = LOW_VOLUME;
	registers[3] = LOW_VOLUME;
	registers[4] = LOW_VOLUME;
	registers[5] = LOW_VOLUME;
	registers[6] = LOW_VOLUME;
	registers[7] = LOW_VOLUME;
	setRegBits(8,7,7,SPDIF_ENABLE);
	setRegBits(8,6,0,AUTOMUTE_TRIGGER_LOWEST);
	registers[9] = AUTOMUTE_OFF;	
	setRegBits(10,7,6,PCM_DATA_LENGTH);
	setRegBits(10,5,4,PCM_DATA_FORMAT);
	setRegBits(10,3,3,NON_OVERLAP_CONTROL);
	setRegBits(10,2,2,JITTER_REDUCTION);
	setRegBits(10,1,1,BYPASS_DEEMPHASIS);
	setRegBits(10,0,0,MUTE);
	setRegBits(11,7,7,MUTE);
	setRegBits(11,6,5,0);					// reserved set to 0b00
	setRegBits(11,4,2,DPLL_BANDWIDTH);
	setRegBits(11,1,0,PCM_DEEMPH_FREQ);
	setRegBits(12,7,7,DITHER_CONTROL);
	setRegBits(12,6,6,ROTATOR_INPUT);
	setRegBits(12,5,5,REMMAPPING);
	setRegBits(12,4,0,NOTCH_DELAY);
	registers[13] = 0;						// all DACs in phase.
	setRegBits(14,7,7,DAC_8_7_4_3_SOURCE);
	setRegBits(14,6,6,DAC_8_7_4_3_SOURCE);
	setRegBits(14,5,5,DAC_8_7_4_3_SOURCE);
	setRegBits(14,4,4,DAC_8_7_4_3_SOURCE);	
	setRegBits(14,3,3,TRUE_DIFFERENTIAL);
	setRegBits(14,2,1,IIR_BANDWIDTH);
	setRegBits(14,0,0,ROLLOFF);
	setRegBits(15,7,6,QUANTIZER_SIZE);
	setRegBits(15,5,4,QUANTIZER_SIZE);
	setRegBits(15,3,2,QUANTIZER_SIZE);
	setRegBits(15,1,0,QUANTIZER_SIZE);
}

void setRegBits(int reg, int hi, int low, int value) {
	int i;
	unsigned char v;
	for (i = low; i <= hi; i++) {
		// clear the register bits in case they were already set
		registers[reg] &= ~(1 << i);
	}
	// shift
	v = (unsigned char) value << low;
	// set
	registers[reg] |= v;
}